/*
 * Copyright (c) 2003-2007 Erez Zadok
 * Copyright (c) 2003-2007 Stony Brook University
 * Copyright (c) 2003-2007 The Research Foundation of State University of New York
 * Copyright (c) 2004-2007 Nikolai Joukov
 * Copyright (c) 2003-2005 Akshat Aranya
 * Copyright (c) 2006-2007 Gopala Surya
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */
/*
 * structures.h
 * This file contains declarations for trace condition tree
 * as maintained by the kernel.
 */

#ifndef __STRUCTURES_H
#define __STRUCTURES_H


#include "messages.h" /* for OP_MAX */
#include "fist.h"
#include "syscalls.h" /* for SYSCALL_OFFSET */
/*
 * This enumeration defines the type of
 * nodes in the trace condition tree
 */
enum node_type {
  TC_TRUE,                 /* TRUE node */
  TC_FALSE,                /* FALSE node */
  TC_OP,                   /* bitmap of F/S ops to be traced */
  TC_SYSCALL,              /* bitmap of system calls to be traced, may not be used */
  TC_CUID,                 /* current UID */
  TC_CGID,                 /* current GID */
  TC_NAME,                 /* File name */
  TC_EXT,                  /* File extension, for example *.c */
  TC_PID,                  /* Process ID */
  TC_PGRP,                 /* Process group */
  TC_FUID,                 /* File owner UID */
  TC_FGID,                 /* File owner GID */
  TC_PNAME,                /* Name of process */
  TC_SID,                  /* Session ID */
  TC_TIME,                 /* Time based */
  TC_INO,                   /* Inode number based */
  TC_MAX_TYPES             /* !!!MUST BE LAST!!! */
};

#ifndef __KERNEL__
#include <sys/time.h>


/* must be kept in sync with the enumeration, w.r.t. indices since we'll use the node_type as an index */
static struct {
    enum node_type t;
    char *name;
} node_names[] = {
    { TC_TRUE, "TC_TRUE" },
    { TC_FALSE, "TC_FALSE" },
    { TC_OP, "TC_OP" },
    { TC_SYSCALL, "TC_SYSCALL" },
    { TC_CUID, "TC_CUID" },
    { TC_CGID, "TC_CGID" },
    { TC_NAME, "TC_NAME" },
    { TC_EXT, "TC_EXT" },
    { TC_PID, "TC_PID" },
    { TC_PGRP, "TC_PGRP" },
    { TC_FUID, "TC_FUID" },
    { TC_FGID, "TC_FGID" },
    { TC_PNAME, "TC_PNAME" },
    { TC_SID, "TC_SID" },
    { TC_TIME, "TC_TIME" },
    { TC_INO, "TC_INO" }
};

static inline char *get_node_type(int t) {
    if (t < TC_MAX_TYPES)
	return node_names[t].name;
    else
	return "UNKNOWN";
}



#endif

struct andor;

/*
 * This structure is for time based tracing
 */
struct time {
  time_t start_time;
  time_t stop_time;
};

/*
 * This structure is for file name, process name based
 * tracing
 */

//struct name {
//  int inum;
  /* TODO: Work out this structure */
//};

#define OP_IS_ENABLED(bitmap, op) ((bitmap)[(op)/(8*sizeof(char))] & 0x1 << ((op)%(8*sizeof(char))))
#define OP_SET_ENABLED(bitmap, op) ((bitmap)[(op)/(8*sizeof(char))] |= 0x1 << ((op)%(8*sizeof(char))))

#define SYSCALL_IS_ENABLED(bitmap, id) \
	(((id) >= SYSCALL_OFFSET) && (((bitmap)[((id) - SYSCALL_OFFSET)/8] & 0x1 << (((id) - SYSCALL_OFFSET)%8))))

#define SYSCALL_MAX 255
#define BITMAP_SIZE  (MAX(OP_MAX, SYSCALL_MAX)/(8*sizeof(char)) + 1)

#ifdef __KERNEL__
struct tracecond {
    enum node_type tc_type;              /* type of this node */
    int tc_num_eval;                     /* number of times this rule was evaluated */
    int tc_num_eval_true;                /* number of times this rule was found to be true */
    struct tracecond *tc_follow_true;
    struct tracecond *tc_follow_false;
    struct list_head tc_list;            /* list of trace conditions, to iterate through the conditions */
    union {
	/* These are pointers used by non-leaf nodes */
	/* These are various data fields used by leaf nodes */
	char tc_map[BITMAP_SIZE];  /* bitmap for TC_OP or TC_SYSCALL or TC_PNAME */
	unsigned tc_ival;                  /* for TC_xUID, TC_xGID, TC_PID, TC_SID, TC_INO, TC_PGRP */
	struct name *tc_name;              /* for TC_NAME and TC_EXT */
	struct time tc_time;              /* for TC_TIME */
    } u;
};

/* Forward declarations for trace_destination, and op vectors. */
struct output_driver;
struct assembly_driver;

#define TD_FREE  	0
#define TD_CONSTRUCTING 1 // transitioning between the other states
#define TD_ODRIVER      2 // Output drivers added
#define TD_ADRIVER      3 // Assembly drivers added
#define TD_DISABLED 	4
#define TD_ENABLED  	5
#define TD_ERROR        6 // error occurred, tracing stopped
/*
 * This structure defines one trace destination
 */
struct trace_destination {
    int td_state;               	/* "state" this destination is in */
    struct tracecond *td_root;    	/* root of the trace condition tree */
    struct assembly_driver *td_ad;	/* first assembly driver in the asse */
    struct output_driver *td_od;  	/* first output driver in the output driver chain */
    struct list_head td_conds;          /* list to store the trace conditions, provides easy mechanism for add/delete */

};

//For write, so we know whether to use memcpy() or copy_from_user()
#define OO_USER_BUFFER   1
#define OO_KERNEL_BUFFER 0

/*
 * Output driver operations
 */
struct output_ops {
	int (*write)(struct output_driver *this, void *buf, int len,int is_user_buffer); /* Write information */
	int (*flush)(struct output_driver *this);                   /* Empty buffered data */
	int (*getblocksize)(struct output_driver *this);            /* Return preferred block size of
								       this driver, but need not be obeyed */
	int (*init)(struct output_driver *this, void *params, struct output_driver *next, int level);       /* Initialize this driver, level is the level in the stack for this output driver with 0=highest */
	void (*release)(struct output_driver *this);                /* end of operation, release resources */
};

struct assembly_ops {
	int (*init)(struct assembly_driver *this, void *options);
	void (*release)(struct assembly_driver *this);
	/* inode ops */
	void (*ao_create_pre)(struct assembly_driver *this, struct trace_destination *dest, struct inode *dir, struct dentry *dentry, int mode, void **context);
	void (*ao_create_post)(struct assembly_driver *this, struct trace_destination *dest, struct inode *dir, struct dentry *dentry, int mode, int ret, void **context);
	void (*ao_lookup_pre)(struct assembly_driver *this, struct trace_destination *dest, struct inode *dir, struct dentry *dentry, void **context);
	//Note: new_dentry for ao_lookup_post is error encoded with ERR_PTR in case of error. It needs to be converted to error code
        //      It is *never* a valid dentry
	void (*ao_lookup_post)(struct assembly_driver *this, struct trace_destination *dest, struct inode *dir, struct dentry *dentry, struct dentry *new_dentry, void **context);
	void (*ao_link_pre)(struct assembly_driver *this, struct trace_destination *dest, struct dentry *old_dentry, struct inode *dir, struct dentry *new_dentry, void **context);
	void (*ao_link_post)(struct assembly_driver *this, struct trace_destination *dest, struct dentry *old_dentry, struct inode *dir, struct dentry *new_dentry, int ret, void **context);
	void (*ao_unlink_pre)(struct assembly_driver *this, struct trace_destination *dest, struct inode *dir, struct dentry *dentry, void **context);
	void (*ao_unlink_post)(struct assembly_driver *this, struct trace_destination *dest, struct inode *dir, struct dentry *dentry, int ret, void **context);
	void (*ao_symlink_pre)(struct assembly_driver *this, struct trace_destination *dest, struct inode *dir, struct dentry *dentry, const char *symname, void **context);
	void (*ao_symlink_post)(struct assembly_driver *this, struct trace_destination *dest, struct inode *dir, struct dentry *dentry, const char *symname, int ret, void **context);
	void (*ao_mkdir_pre)(struct assembly_driver *this, struct trace_destination *dest, struct inode *dir, struct dentry *dentry, int mode, void **context);
	void (*ao_mkdir_post)(struct assembly_driver *this, struct trace_destination *dest, struct inode *dir, struct dentry *dentry, int mode, int ret, void **context);
	void (*ao_rmdir_pre)(struct assembly_driver *this, struct trace_destination *dest, struct inode *dir, struct dentry *dentry, void **context);
	void (*ao_rmdir_post)(struct assembly_driver *this, struct trace_destination *dest, struct inode *dir, struct dentry *dentry, int ret, void **context);
	void (*ao_mknod_pre)(struct assembly_driver *this, struct trace_destination *dest, struct inode *dir, struct dentry *denry, int mode, unsigned long dev, void **context);
	void (*ao_mknod_post)(struct assembly_driver *this, struct trace_destination *dest, struct inode *dir, struct dentry *denry, int mode, unsigned long dev, int ret, void **context);
	void (*ao_rename_pre)(struct assembly_driver *this, struct trace_destination *dest, struct inode *old_dir, struct dentry *old_dentry, struct inode *new_dir, struct dentry *new_dentry, void **context);
	void (*ao_rename_post)(struct assembly_driver *this, struct trace_destination *dest, struct inode *old_dir, struct dentry *old_dentry, struct inode *new_dir, struct dentry *new_dentry, int ret, void **context);
	void (*ao_readlink_pre)(struct assembly_driver *this, struct trace_destination *dest, struct dentry *dentry, char *buf, int bufsiz, void **context);
	void (*ao_readlink_post)(struct assembly_driver *this, struct trace_destination *dest, struct dentry *dentry, char *buf, int bufsiz, int ret, void **context);
	void (*ao_follow_link_pre)(struct assembly_driver *this, struct trace_destination *dest, struct dentry *dentry, struct nameidata *nd, void **context);
	void (*ao_follow_link_post)(struct assembly_driver *this, struct trace_destination *dest, struct dentry *dentry, struct nameidata *nd, int ret, void **context);
	void (*ao_permission_pre)(struct assembly_driver *this, struct trace_destination *dest, struct inode *inode,int mask, void **context);
	void (*ao_permission_post)(struct assembly_driver *this, struct trace_destination *dest, struct inode *inode,int mask, int ret, void **context);
	void (*ao_inode_revalidate_pre)(struct assembly_driver *this, struct trace_destination *dest, struct dentry *dentry, void **context);
	void (*ao_inode_revalidate_post)(struct assembly_driver *this, struct trace_destination *dest, struct dentry *dentry, int ret, void **context);
	void (*ao_setattr_pre)(struct assembly_driver *this, struct trace_destination *dest, struct dentry *dentry, struct iattr *ia, void **context);
	void (*ao_setattr_post)(struct assembly_driver *this, struct trace_destination *dest, struct dentry *dentry, struct iattr *ia, int ret, void **context);
	/* file ops */
	void (*ao_llseek_pre)(struct assembly_driver *this, struct trace_destination *dest, struct file *file, loff_t offset, int origin, void **context);
	void (*ao_llseek_post)(struct assembly_driver *this, struct trace_destination *dest, struct file *file, loff_t offset, int origin, loff_t ret, void **context);
	void (*ao_read_pre)(struct assembly_driver *this, struct trace_destination *dest, file_t *file, char *buf, size_t count, loff_t *ppos, void **context);
	void (*ao_read_post)(struct assembly_driver *this, struct trace_destination *dest, file_t *file, char *buf, size_t count, loff_t *ppos, ssize_t ret, void **context);
	void (*ao_write_pre)(struct assembly_driver *this, struct trace_destination *dest, file_t *file, const char *buf, size_t count, loff_t *ppos, void **context);
	void (*ao_write_post)(struct assembly_driver *this, struct trace_destination *dest, file_t *file, const char *buf, size_t count, loff_t *ppos, ssize_t ret, void **context);
	void (*ao_readdir_pre)(struct assembly_driver *this, struct trace_destination *dest, file_t *file, void *dirent, filldir_t filldir, int count, void **context);
	void (*ao_readdir_post)(struct assembly_driver *this, struct trace_destination *dest, file_t *file, void *firent, filldir_t filldir, int count, int ret, void **context);
	void (*ao_poll_pre)(struct assembly_driver *this, struct trace_destination *dest, file_t *file, poll_table *wait, void **context);
	void (*ao_poll_post)(struct assembly_driver *this, struct trace_destination *dest, file_t *file, poll_table *wait, int ret, void **context);
	void (*ao_ioctl_pre)(struct assembly_driver *this, struct trace_destination *dest, inode_t *inode, file_t *file, unsigned int cmd, unsigned long arg, void **context);
	void (*ao_ioctl_post)(struct assembly_driver *this, struct trace_destination *dest, inode_t *inode, file_t *file, unsigned int cmd, unsigned long arg, int ret, void **context);
	void (*ao_mmap_pre)(struct assembly_driver *this, struct trace_destination *dest, file_t *file, vm_area_t *vma, void **context);
	void (*ao_mmap_post)(struct assembly_driver *this, struct trace_destination *dest, file_t *file, vm_area_t *vma, int ret, void **context);
	void (*ao_open_pre)(struct assembly_driver *this, struct trace_destination *dest, inode_t *inode, file_t *file, void **context);
	void (*ao_open_post)(struct assembly_driver *this, struct trace_destination *dest, inode_t *inode, file_t *file, int ret, void **context);
	void (*ao_flush_pre)(struct assembly_driver *this, struct trace_destination *dest, file_t *file, void **context);
	void (*ao_flush_post)(struct assembly_driver *this, struct trace_destination *dest, file_t *file, int ret, void **context);
	void (*ao_release_pre)(struct assembly_driver *this, struct trace_destination *dest, inode_t *inode, file_t *file, void **context);
	void (*ao_release_post)(struct assembly_driver *this, struct trace_destination *dest, inode_t *inode, file_t *file, int ret, void **context);
	void (*ao_fsync_pre)(struct assembly_driver *this, struct trace_destination *dest, file_t *file, dentry_t *dentry, int datasync, void **context);
	void (*ao_fsync_post)(struct assembly_driver *this, struct trace_destination *dest, file_t *file, dentry_t *dentry, int datasync, int ret, void **context);
	void (*ao_fasync_pre)(struct assembly_driver *this, struct trace_destination *dest, int fd, file_t *file, int flag, void **context);
	void (*ao_fasync_post)(struct assembly_driver *this, struct trace_destination *dest, int fd, file_t *file, int flag, int ret, void **context);
	void (*ao_lock_pre)(struct assembly_driver *this, struct trace_destination *dest, file_t *file, int cmd, struct file_lock *fl, void **context);
	void (*ao_lock_post)(struct assembly_driver *this, struct trace_destination *dest, file_t *file, int cmd, struct file_lock *fl, int ret, void **context);
	/* super ops */
	void (*ao_read_inode_pre)(struct assembly_driver *this, struct trace_destination *dest, inode_t *inode, void **context);
	void (*ao_read_inode_post)(struct assembly_driver *this, struct trace_destination *dest, inode_t *inode, void **context);
	void (*ao_write_inode_pre)(struct assembly_driver *this, struct trace_destination *dest, inode_t *inode, int sync, void **context);
	void (*ao_write_inode_post)(struct assembly_driver *this, struct trace_destination *dest, inode_t *inode, int sync, void **context);
	void (*ao_put_inode_pre)(struct assembly_driver *this, struct trace_destination *dest, inode_t *inode, void **context);
	void (*ao_put_inode_post)(struct assembly_driver *this, struct trace_destination *dest, inode_t *inode, void **context);
	void (*ao_delete_inode_pre)(struct assembly_driver *this, struct trace_destination *dest, inode_t *inode, void **context);
	void (*ao_delete_inode_post)(struct assembly_driver *this, struct trace_destination *dest, inode_t *inode, void **context);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	void (*ao_statfs_pre)(struct assembly_driver *this, struct trace_destination *dest, struct statfs *buf, void **context);
	void (*ao_statfs_post)(struct assembly_driver *this, struct trace_destination *dest, struct statfs *buf, int ret, void **context);
#else
	void (*ao_statfs_pre)(struct assembly_driver *this, struct trace_destination *dest, struct kstatfs *buf, void **context);
	void (*ao_statfs_post)(struct assembly_driver *this, struct trace_destination *dest, struct kstatfs *buf, int ret, void **context);
#endif
	void (*ao_clear_inode_pre)(struct assembly_driver *this, struct trace_destination *dest, inode_t *inode, void **context);
	void (*ao_clear_inode_post)(struct assembly_driver *this, struct trace_destination *dest, inode_t *inode, void **context);
	/* dentry ops */
	void (*ao_d_revalidate_pre)(struct assembly_driver *this, struct trace_destination *dest, dentry_t *dentry, int flags, void **context);
	void (*ao_d_revalidate_post)(struct assembly_driver *this, struct trace_destination *dest, dentry_t *dentry, int flags, int ret, void **context);
	void (*ao_d_hash_pre)(struct assembly_driver *this, struct trace_destination *dest, dentry_t *dentry, qstr_t *name, void **context);
	void (*ao_d_hash_post)(struct assembly_driver *this, struct trace_destination *dest, dentry_t *dentry, qstr_t *name, int ret, void **context);
	void (*ao_d_compare_pre)(struct assembly_driver *this, struct trace_destination *dest, dentry_t *dentry, qstr_t *a, qstr_t *b, void **context);
	void (*ao_d_compare_post)(struct assembly_driver *this, struct trace_destination *dest, dentry_t *dentry, qstr_t *a, qstr_t *b, int ret, void **context);
	void (*ao_d_delete_pre)(struct assembly_driver *this, struct trace_destination *dest, dentry_t *dentry, void **context);
	void (*ao_d_delete_post)(struct assembly_driver *this, struct trace_destination *dest, dentry_t *dentry, int ret, void **context);
	void (*ao_d_release_pre)(struct assembly_driver *this, struct trace_destination *dest, dentry_t *dentry, void **context);
	void (*ao_d_release_post)(struct assembly_driver *this, struct trace_destination *dest, dentry_t *dentry, void **context);
	void (*ao_d_iput_pre)(struct assembly_driver *this, struct trace_destination *dest, dentry_t *dentry, inode_t *inode, void **context);
	void (*ao_d_iput_post)(struct assembly_driver *this, struct trace_destination *dest, dentry_t *dentry, inode_t *inode, void **context);
	void (*ao_readpage_pre)(struct assembly_driver *this, struct trace_destination *dest, struct file *file, struct page *page, void **context);
	void (*ao_readpage_post)(struct assembly_driver *this, struct trace_destination *dest, struct file *file, struct page *page, int ret, void **context);
	void (*ao_writepage_pre)(struct assembly_driver *this, struct trace_destination *dest, struct page *page, void **context);
	void (*ao_writepage_post)(struct assembly_driver *this, struct trace_destination *dest, struct page *page, int ret, void **context);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	void (*ao_readpages_pre)(struct assembly_driver *this, struct trace_destination *dest, struct file *file, struct address_space* mapping, struct list_head *pages, void **context);
	void (*ao_readpages_post)(struct assembly_driver *this, struct trace_destination *dest, struct file *file, struct address_space* mapping, struct list_head *pages, int ret, void **context);
#endif

	void (*ao_prepare_write_pre)(struct assembly_driver *this, struct trace_destination *dest, struct file *file, struct page *page, unsigned from, unsigned to, void **context);
	void (*ao_prepare_write_post)(struct assembly_driver *this, struct trace_destination *dest, struct file *file, struct page *page, unsigned from, unsigned to, int ret, void **context);
	void (*ao_commit_write_pre)(struct assembly_driver *this, struct trace_destination *dest, struct file *file, struct page *page, unsigned from, unsigned to, void **context);
	void (*ao_commit_write_post)(struct assembly_driver *this, struct trace_destination *dest, struct file *file, struct page *page, unsigned from, unsigned to, int ret, void **context);
        void (*ao_sync_page_pre)(struct assembly_driver *this,struct trace_destination *dest,page_t* page,void** context);
        void (*ao_sync_page_post)(struct assembly_driver *this,struct trace_destination *dest,page_t* page,int ret,void** context);



	void (*ao_start_trace)(struct assembly_driver *this, struct trace_destination *dest);
	void (*ao_stop_trace)(struct assembly_driver *this, struct trace_destination *dest);

};

/*
 * This structure is hung off the private data in the superblock
 * of the mounted file system.
 */

struct output_driver {
	int od_type; // OD_COMPRESS, OD_CKSUM, OD_ENCRYPT
	            // OD_FILE, OD_SOCKET, OD_NETLINK, OD_RAW ...
	void *od_private_data; // Store driver specific parameters
	struct output_ops *od_ops; /* output ops vector */
	struct output_driver *od_next; /* next driver in chain */
};


struct assembly_driver {
	int ad_type; // AD_*
	void *ad_private_data;
	struct assembly_ops *ad_ops;
	struct assembly_driver *ad_next;
   	struct semaphore ad_odriver_sem;
};


/*
 * This structure is hung off the private data of the super block
 * of the mounted file system
 */
#define MAX_DESTINATIONS 10          /* Total number of possible destinations */
struct trace_sb {
    struct trace_destination dest[MAX_DESTINATIONS];
};

#endif

#endif /* __STRUCTURES_H */
